(*============================================================================*)
(* CHESS.PAS -- Chess object                                                  *)
(*----------------------------------------------------------------------------*)
(*                                                                            *)
(* BPGNVIEW.EXE -- BPGN file viewer.                                          *)
(* Copyright (C) 1998  Zon Juan Jan, 2002-2005 Fermy                          *)
(*                                                                            *)
(* This program is free software; you can redistribute it and/or              *)
(* modify it under the terms of the GNU General Public License                *)
(* as published by the Free Software Foundation; either version 2             *)
(* of the License, or (at your option) any later version.                     *)
(*                                                                            *)
(* This program is distributed in the hope that it will be useful,            *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of             *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *)
(* GNU General Public License for more details.                               *)
(*                                                                            *)
(* You should have received a copy of the GNU General Public License          *)
(* along with this program; if not, write to the Free Software Foundation,    *)
(* Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.            *)
(*                                                                            *)
(*----------------------------------------------------------------------------*)
(* HISTORY                                                                    *)
(* -------                                                                    *)
(* 1998.11.03: First construction.                                            *)
(* 1998.11.09: Added GetAlgebraicMoveStr().                                   *)
(* 1998.11.11: Added some parameters to ExecAlgebraicMove().                  *)
(* 1999.01.20: 1) Added code to read in Crazyhouse games (determined by event *)
(*               tag in BPGN file: [Event "fics unrated crazyhouse match"]).  *)
(*               If it's a crazyhouse game then the captured piece's color    *)
(*               will be reversed and into player's piece buffer instead of   *)
(*               going to player partner's piece buffer as in regular         *)
(*               bughouse games.                                              *)
(*            2) Changed version ID to 1.13 beta.                             *)
(* 2001.12.12: Added routine AlgebraicMove that converts MoveRecord -> Text   *)
(*             move, e.g. n@f3                                                *)
(* 2002.04.11: Added routine setfenpos for Fen parsing                        *)
(* 2004.11.16: Added tchess.addDropPiece                                      *)
(* 2004.11.28: Modified ExecAlgebraicMove to check for the value of bairdrop  *)
(* 2004.12.20: Fixed a bug in ExecAlgebraicMove and set pieces that caused    *)
(*           false illegal drop errors with Options-->bpgn-->airdrop unchecked*)
(*============================================================================*)
UNIT CHESS;

INTERFACE

USES SysUtils,global,misc;


TYPE
    TChess = CLASS(TObject)
        CONSTRUCTOR Init;
        DESTRUCTOR  Destroy;

        PROCEDURE   ResetBoard;
        PROCEDURE   SetPieces(cWPc, cBPc: String);
        PROCEDURE   SetPosition(cSt, cWPc, cBPc: String);
        procedure   setfenpos(st:string);

        FUNCTION    GetIndex(cPc: Char): Integer;
        FUNCTION    GetRevIndex(nBI: Integer): Integer;
        FUNCTION    IndexOk(nI: Integer): Boolean;

        FUNCTION    GetPiece(nPc, nIdx: Integer): Char;
        FUNCTION    GetRevPiece(cPc: Char): Char;
        FUNCTION    GetColor(cPc: Char): Integer;
        FUNCTION    GetRevColor(cPc: Char): Integer;

        FUNCTION    GetWhiteColor: Integer;
        FUNCTION    GetBlackColor: Integer;

        FUNCTION    GetWhitePawn: Char;
        FUNCTION    GetWhiteRook: Char;
        FUNCTION    GetWhiteKnight: Char;
        FUNCTION    GetWhiteBishop: Char;
        FUNCTION    GetWhiteQueen: Char;
        FUNCTION    GetWhiteKing: Char;

        FUNCTION    GetBlackPawn: Char;
        FUNCTION    GetBlackRook: Char;
        FUNCTION    GetBlackKnight: Char;
        FUNCTION    GetBlackBishop: Char;
        FUNCTION    GetBlackQueen: Char;
        FUNCTION    GetBlackKing: Char;

        FUNCTION    GetEmpty: Char;

        FUNCTION    GetPawnIndex: Integer;
        FUNCTION    GetRookIndex: Integer;
        FUNCTION    GetKnightIndex: Integer;
        FUNCTION    GetBishopIndex: Integer;
        FUNCTION    GetQueenIndex: Integer;
        FUNCTION    GetKingIndex: Integer;

        FUNCTION    IsWhite(cPc: Char): Boolean;
        FUNCTION    IsBlack(cPc: Char): Boolean;

        FUNCTION    IsPawn(cPc: Char): Boolean;
        FUNCTION    IsRook(cPc: Char): Boolean;
        FUNCTION    IsKnight(cPc: Char): Boolean;
        FUNCTION    IsBishop(cPc: Char): Boolean;
        FUNCTION    IsQueen(cPc: Char): Boolean;
        FUNCTION    IsKing(cPc: Char): Boolean;

        FUNCTION    IsWhitePawn(cPc: Char): Boolean;
        FUNCTION    IsWhiteRook(cPc: Char): Boolean;
        FUNCTION    IsWhiteKnight(cPc: Char): Boolean;
        FUNCTION    IsWhiteBishop(cPc: Char): Boolean;
        FUNCTION    IsWhiteQueen(cPc: Char): Boolean;
        FUNCTION    IsWhiteKing(cPc: Char): Boolean;

        FUNCTION    IsBlackPawn(cPc: Char): Boolean;
        FUNCTION    IsBlackRook(cPc: Char): Boolean;
        FUNCTION    IsBlackKnight(cPc: Char): Boolean;
        FUNCTION    IsBlackBishop(cPc: Char): Boolean;
        FUNCTION    IsBlackQueen(cPc: Char): Boolean;
        FUNCTION    IsBlackKing(cPc: Char): Boolean;

        FUNCTION    IsEmpty(cPc: Char): Boolean;
        FUNCTION    IsPromoteRank(nPc, nY: Integer): Boolean;

        PROCEDURE   AddMove(nFromX, nFromY, nToX, nToY: Integer;
                            cPiece, cPromote, cType: Char);

        PROCEDURE   GenPawnMove  (cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
        PROCEDURE   GenCrossMove (cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
        PROCEDURE   GenXShapeMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
        PROCEDURE   GenKnightMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
        PROCEDURE   GenKingMove  (cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);

        PROCEDURE   GenPieceMove(bWM: Boolean);
        PROCEDURE   GenOnePieceMove(bWM: Boolean; cPc: Char);
        PROCEDURE   GenDropMove (bWM: Boolean);
        PROCEDURE   GenOneDropMove(bWM: Boolean; cPc: Char);
        PROCEDURE   GenMove;

        PROCEDURE   BackupSetting;
        PROCEDURE   RestoreSetting;

        FUNCTION    SquareAttacked(nFromX, nFromY, nPc: Integer): Boolean;

        FUNCTION    SmartMove(nFromX, nFromY: Integer): Integer;
        FUNCTION    MatchMove(nFromX, nFromY, nToX, nToY: Integer;
                              cPiece, cPromote: Char): Integer;

        FUNCTION    LegalMove(nFromX, nFromY, nToX, nToY: Integer;
                              cPiece, cPromote: Char): Integer;

        FUNCTION    ExecMove(nMI: Integer): Integer;

        FUNCTION    GetBoardStr: String;
        FUNCTION    GetWhitePieceStr: String;
        FUNCTION    GetBlackPieceStr: String;
        FUNCTION    GetFENBoardStr: String;
        FUNCTION    GetMoveStr(nMI: Integer; bCheckKing, bMateKing: Boolean): String;
        PROCEDURE   GetAlgebraicMoveStr(nMI: Integer; bCheckKing, bMateKing: Boolean; VAR cM1, cM2, cM3, cM4: String);
        FUNCTION    GetLongMoveStr(nMI: Integer): String;
        FUNCTION    ExecAlgebraicMove(nMoveNum: Integer; cBoard: Char; cOrgMove: String; VAR nMI: Integer): String;
        FUNCTION    GetRealMove(nMI: Integer; cOrgMove: String): String;
        procedure   AddDropPiece(cpc:char);
        procedure   GoClear;
        procedure   undoSetup;
        function    getEnpassantsq:string;
        function    GetBFenStr:string;
        PRIVATE
            { Backup & restore }
            aTmpKingBX:     ARRAY[1..2] OF Integer;
            aTmpKingBY:     ARRAY[1..2] OF Integer;

            aTmpKCastle:    ARRAY[1..2] OF Boolean;
            aTmpQCastle:    ARRAY[1..2] OF Boolean;

            mTmpPrev:       MoveRecord;

            aTmpBoard:      BoardArray;
            aTmpPromote:    PromoteArray;
            aTmpcPiece:     ARRAY[1..2,1..8] OF Char;
            aTmpnPiece:     ARRAY[1..2,1..8] OF Integer;

            bTmpWhitesMove: Boolean;

        PUBLIC
            { Current }
            aKingBX:     ARRAY[1..2] OF Integer;  // kings' file positions [1]= white king file
            aKingBY:     ARRAY[1..2] OF Integer;  // ------ rank ---------[1]= white king rank

            aKCastle:    ARRAY[1..2] OF Boolean;      //  castling possibilities. [1] for white
            aQCastle:    ARRAY[1..2] OF Boolean;      //

            mPrev:       MoveRecord;

            nMove:       Integer;
            aMove:       ARRAY[1..MAX_MOVE] OF MoveRecord;
{looks like this array stores all possible moves in a given position, nMove is number of those.}

            aBoard:      BoardArray;                 // [8,8] array. Translation to string[64] array:
                                                        //st[1] -> a1->ba[1,1], st[64]->h8->ba[8,8]
                                                        // st[8] ->h1->ba[8,1]
            aPromote:    PromoteArray;
            acPiece:     ARRAY[1..2,1..8] OF Char;    //holding, just names of pieces (Q,q,n,etc)
                                                        // [1,x] are whites, [2,x] are blacks
                                                        // [x,1] are p's [x,2]=n, then b,r,q.
            anPiece:     ARRAY[1..2,1..8] OF Integer;  //number of each piece in holding
            bWhitesMove: Boolean;
            cSaveBPieces:string;
            cSaveWPieces:string;
            cSaveBoard:String;

    END;

VAR
    { *** CROSS *** }
    aCrossX:  ARRAY[1..MAX_DIR, 1..MAX_DIST] OF Integer =
              (( 0,  0,  0,  0,  0,  0,  0),
               ( 0,  0,  0,  0,  0,  0,  0),
               (+1, +2, +3, +4, +5, +6, +7),
               (-1, -2, -3, -4, -5, -6, -7));
    aCrossY:  ARRAY[1..MAX_DIR, 1..MAX_DIST] OF Integer =
              ((+1, +2, +3, +4, +5, +6, +7),
               (-1, -2, -3, -4, -5, -6, -7),
               ( 0,  0,  0,  0,  0,  0,  0),
               ( 0,  0,  0,  0,  0,  0,  0));

    { *** X SHAPE *** }
    aXShapeX: ARRAY[1..MAX_DIR, 1..MAX_DIST] OF Integer =
              ((+1, +2, +3, +4, +5, +6, +7),
               (-1, -2, -3, -4, -5, -6, -7),
               (+1, +2, +3, +4, +5, +6, +7),
               (-1, -2, -3, -4, -5, -6, -7));
    aXShapeY: ARRAY[1..MAX_DIR, 1..MAX_DIST] OF Integer =
              ((+1, +2, +3, +4, +5, +6, +7),
               (-1, -2, -3, -4, -5, -6, -7),
               (-1, -2, -3, -4, -5, -6, -7),
               (+1, +2, +3, +4, +5, +6, +7));

    { *** KNIGHT *** }
    aKnightX: ARRAY[1..MAX_KNIGHT] OF Integer =
              (+1, +2, +2, +1, -1, -2, -2, -1);
    aKnightY: ARRAY[1..MAX_KNIGHT] OF Integer =
              (+2, +1, -1, -2, -2, -1, +1, +2);

    { *** KING *** }
    aKingX: ARRAY[1..MAX_KING] OF Integer =
            ( 0, +1, +1, +1,  0, -1, -1, -1);
    aKingY: ARRAY[1..MAX_KING] OF Integer =
            (+1, +1,  0, -1, -1, -1,  0, +1);

    { *** PAWN PROMOTE Y *** }
    aPromoteY: ARRAY[1..2] OF Integer = (8, 1);

    { *** ORIGINAL ROOK Y POSITION *** }
    aOrgRookY: ARRAY[1..2] OF Integer = (1, 8);

    { *** PAWN GENERATION *** }
    aGenPawnAdvY: ARRAY[1..2] OF Integer = (1, -1);
    aGenPawnOrgY: ARRAY[1..2] OF Integer = (2, 7);
    aGenPawnEPY:  ARRAY[1..2] OF Integer = (5, 4);

    { *** KING CASTLE *** }
    aOrgKingY: ARRAY[1..2] OF Integer = (1, 8);

    { *** WHITE DROP PIECES *** }
    aDropPiece: ARRAY[1..2, 1..6] OF Char =
                ((WHITE_PAWN, WHITE_ROOK, WHITE_KNIGHT, WHITE_BISHOP, WHITE_QUEEN, WHITE_KING),
                 (BLACK_PAWN, BLACK_ROOK, BLACK_KNIGHT, BLACK_BISHOP, BLACK_QUEEN, BLACK_KING));

function AlgebraicMove(rec:MoveRecord):string;

IMPLEMENTATION

CONSTRUCTOR TChess.Init;
BEGIN
    ResetBoard;
END;

DESTRUCTOR TChess.Destroy;
BEGIN
    { Nothing }
END;

procedure Tchess.setfenpos(st:string);
var ps,i,j,k,l,h,m,er:integer;
s,s1,pieces,fen,files:string;
begin
 fen:=st;
 pieces:='KQRNBPkqrnbp';
 files:='abcdefgh';
 ps:=pos(' ',st);
 if (ps>0) then
 begin s:=copy(st,1,ps-1);delete(st,1,ps);end
 else begin s:=st;st:='';end;
 s:=s+'/';
 for i:=1 to 8 do
 begin
  k:=0;
  ps:=pos('/',s);
  if (ps<0) then
  begin
   resetboard;
   wError('Invalid FEN tag: missing rank(s). FEN='+fen);
   exit;
  end;     // missing rank
  s1:=copy(s,1,ps-1);
  delete(s,1,ps);
  l:=length(s1);
  for j:=1 to l do
  begin
   if isnumber(s1[j]) then
   begin
    val(s1[j],m,er);
    if (k+m>8) then
    begin
     str(i,s);
     wError('Invalid FEN tag: Rank#'+s+' is too long. FEN='+fen);
     resetboard;exit;
    end;         // rank is too long
    for h:=1 to m do aBoard[k+h,9-i]:=EMPTY_SQUARE;
    inc(k,m);
    continue;
   end;

   if ((s1[j]='~') and (k>0) and (aBoard[k,9-i]<>EMPTY_SQUARE)) then
   begin
    aPromote[k,9-i]:=true;
    continue;
   end;

   if ((k<8) and (pos(s1[j],pieces)>0)) then
   begin
    inc(k,1);
    aBoard[k,9-i]:=s1[j];
    if s1[j]=WHITE_KING then BEGIN aKingBX[1] := k; aKingBY[1] := 9-i; END;
    if s1[j]=BLACK_KING then BEGIN aKingBX[2] := k; aKingBY[2] := 9-i; END;
   end
   else begin
    if k<8 then s:='Illegal character "'+s1[j]+'". Fen='+fen
    else begin str(i,s); s:='Rank#'+s+' is too long. FEN='+fen;end;
    Werror('Invalid FEN tag: '+s);
    resetboard;exit;
   end;  //invalid character in bfen or rank is too long
  end;
  if k<>8 then
  begin
   str(i,s);
   Werror('Invalid FEN tag: Rank#'+s+' is too short. FEN='+fen);
   resetboard;exit;
  end; //invalid size of the rank: too short;
 end;

 aKcastle[1]:=false;
 aQcastle[1]:=false;
 aKcastle[2]:=false;
 aQcastle[2]:=false;
 bWhitesMove:=True;

 if (length(s)>1) then setPieces(s,s);

 st:=trim(st);
 st:=st+' ';
 ps:=pos(' ',st);
 s:=copy(st,1,ps-1);
 delete(st,1,ps);
 if length(s)<>1 then exit; // expecting side to move;
 if s[1]='b' then bWhitesMove:=False;

 st:=trim(st);
 st:=st+' ';
 ps:=pos(' ',st);
 s:=copy(st,1,ps-1);
 delete(st,1,ps);
 if ((length(s)<1) or (length(s)>4)) then exit;// expecting castling stuff

 if pos('K',s)>0  then aKcastle[1]:=true;
 if pos('k',s)>0  then aKcastle[2]:=true;
 if pos('Q',s)>0  then aQCastle[1]:=true;
 if pos('q',s)>0  then aqcastle[2]:=true;

 st:=trim(st);
 st:=st+' ';
 ps:=pos(' ',st);
 s:=copy(st,1,ps-1);
 if ((length(s)<>2) or ((s[2]<>'3')and (s[2]<>'6'))) then exit; //expecting enpassant square;
 if (bWhitesMove) then
 begin
  if s[2]<>'6' then exit;
  ps:=pos(s[1],files);
  if ps<0 then exit;
  if ((aBoard[ps,5]<>'p') or (aBoard[ps,7]<>EMPTY_SQUARE)) then exit;
  mprev.nFromX:=ps;
  mprev.nFromY:=7;
  mprev.nTox:=ps;
  mprev.ntoy:=5;
  mprev.cpiece:=BLACK_PAWN;
 end
 else begin
  if s[2]<>'3' then exit;
  ps:=pos(s[1],files);
  if ps<0 then exit;
  if ((aBoard[ps,4]<>'P') or (aBoard[ps,2]<>EMPTY_SQUARE)) then exit;
  mprev.nFromX:=ps;
  mprev.nFromY:=2;
  mprev.nTox:=ps;
  mprev.ntoy:=4;
  mprev.cpiece:=WHITE_PAWN;
 end;
end;

function AlgebraicMove(rec:MoveRecord):string;
var s1,tmp,res,r2,r4:string;
begin
 tmp:='abcdefgh';
 if ((rec.nFromY=0) or (rec.NfromX=0)) then s1:=rec.cpiece+'@'
 else begin str(rec.nFromY,r2); s1:=tmp[rec.nFromX]+r2; end;
 str(rec.nToY,r4);
 s1:=s1+tmp[rec.nToX]+r4;
 if (length(trim(rec.cpromote))>0)
 then s1:=s1+rec.cpromote;
 Result:=s1;
end;

PROCEDURE TChess.ResetBoard;
VAR
    nI: Integer;
BEGIN
    SetPosition(INIT_GAME_POS, INIT_WHITE_PC, INIT_BLACK_PC);

    bWhitesMove := TRUE;

    nMove  := 0;
    FOR nI := 1 TO MAX_MOVE DO
    BEGIN
        aMove[nI].nFromX                   := 0;
        aMove[nI].nFromY                   := 0;
        aMove[nI].nToX                     := 0;
        aMove[nI].nToY                     := 0;
        aMove[nI].cPiece                   := EMPTY_SQUARE;
        aMove[nI].cPromote                 := EMPTY_SQUARE;
        aMove[nI].cCapture                 := EMPTY_SQUARE;
        aMove[nI].cType                    := MOVE_REGULAR;
        aMove[nI].bCapturedIsPromotedPiece := FALSE;
    END;

    mPrev.nFromX                   := 0;
    mPrev.nFromY                   := 0;
    mPrev.nToX                     := 0;
    mPrev.nToY                     := 0;
    mPrev.cPiece                   := EMPTY_SQUARE;
    mPrev.cPromote                 := EMPTY_SQUARE;
    mPrev.cCapture                 := EMPTY_SQUARE;
    mPrev.cType                    := MOVE_REGULAR;
    mPrev.bCapturedIsPromotedPiece := FALSE;
END;

PROCEDURE TChess.SetPieces(cWPc, cBPc: String);
VAR
    nI, nJ: Integer;
BEGIN
    { Reset piece buffers }
    FOR nI := 1 TO 2 DO
    BEGIN
        FOR nJ := 1 TO 8 DO
        BEGIN
            acPiece[nI, nJ] := EMPTY_SQUARE;
           anPiece[nI, nJ] := 0;
        END;
   END;

    { Load white piece buffer }
    FOR nI := 1 TO Length(cWPc) DO
    BEGIN
        IF (cWPc[nI] = WHITE_PAWN) THEN
        BEGIN
            acPiece[1, pawn_index] := cWPc[nI];
            anPiece[1, pawn_index] := anPiece[1, pawn_index] + 1;
            CONTINUE;
        END;

        IF (cWPc[nI] = WHITE_KNIGHT) THEN
        BEGIN
            acPiece[1, knight_index] := cWPc[nI];
            anPiece[1, knight_index] := anPiece[1, knight_index] + 1;
            CONTINUE;
        END;

        IF (cWPc[nI] = WHITE_BISHOP) THEN
        BEGIN
            acPiece[1, bishop_index] := cWPc[nI];
            anPiece[1, bishop_index] := anPiece[1,bishop_index] + 1;
            CONTINUE;
        END;

        IF (cWPc[nI] = WHITE_ROOK) THEN
        BEGIN
            acPiece[1, rook_index] := cWPc[nI];
            anPiece[1, rook_index] := anPiece[1, rook_index] + 1;
            CONTINUE;
        END;

        IF (cWPc[nI] = WHITE_QUEEN) THEN
        BEGIN
            acPiece[1, queen_index] := cWPc[nI];
            anPiece[1, queen_index] := anPiece[1, queen_index] + 1;
        END;
    END;

    { Load black piece buffer }
    FOR nI := 1 TO Length(cBPc) DO
    BEGIN
        IF (cBPc[nI] = BLACK_PAWN) THEN
        BEGIN
            acPiece[2, Pawn_index] := cBPc[nI];
            anPiece[2, pawn_index] := anPiece[2, pawn_index] + 1;
            CONTINUE;
        END;

        IF (cBPc[nI] = BLACK_KNIGHT) THEN
        BEGIN
            acPiece[2, knight_index] := cBPc[nI];
            anPiece[2, knight_index] := anPiece[2, knight_index] + 1;
            CONTINUE;
        END;

        IF (cBPc[nI] = BLACK_BISHOP) THEN
        BEGIN
            acPiece[2, bishop_index] := cBPc[nI];
            anPiece[2, bishop_index] := anPiece[2, bishop_index] + 1;
            CONTINUE;
        END;

        IF (cBPc[nI] = BLACK_ROOK) THEN
        BEGIN
            acPiece[2, rook_index] := cBPc[nI];
            anPiece[2, rook_index] := anPiece[2, rook_index] + 1;
            CONTINUE;
        END;

        IF (cBPc[nI] = BLACK_QUEEN) THEN
        BEGIN
            acPiece[2, queen_index] := cBPc[nI];
            anPiece[2, queen_index] := anPiece[2, queen_index] + 1;
        END;
    END;
END;

PROCEDURE TChess.SetPosition(cSt, cWPc, cBPc: String);
VAR
    nI, nX, nY: Integer;
BEGIN
    IF (Length(cSt) < 64) THEN
        EXIT;

    aKingBX[1] := 0;
    aKingBY[1] := 0;

    aKingBX[2] := 0;
    aKingBY[2] := 0;

    nI := 0;
    FOR nY := 1 TO 8 DO
    BEGIN
        FOR nX := 1 TO 8 DO
        BEGIN
            nI := nI + 1;
            aBoard  [nX, nY] := cSt[nI];
            aPromote[nX, nY] := FALSE;

            CASE (aBoard[nX, nY]) OF
                WHITE_KING:
                    BEGIN
                        aKingBX[1] := nX;
                        aKingBY[1] := nY;
                    END;

                BLACK_KING:
                    BEGIN
                        aKingBX[2] := nX;
                        aKingBY[2] := nY;
                    END;
            END;
        END;
    END;

    SetPieces(cWPc, cBPc);

    aKCastle[1] := ((aKingBX[1]  = 5) AND
                    (aKingBY[1]  = 1) AND
                    (aBoard[8,1] = WHITE_ROOK));

    aKCastle[2] := ((aKingBX[2]  = 5) AND
                    (aKingBY[2]  = 8) AND
                    (aBoard[8,8] = BLACK_ROOK));

    aQCastle[1] := ((aKingBX[1]  = 5) AND
                    (aKingBY[1]  = 1) AND
                    (aBoard[1,1] = WHITE_ROOK));

    aQCastle[2] := ((aKingBX[2]  = 5) AND
                    (aKingBY[2]  = 8) AND
                    (aBoard[1,8] = BLACK_ROOK));
END;

FUNCTION TChess.GetIndex(cPc: Char): Integer;
BEGIN
    CASE (cPc) OF
        WHITE_PAWN:   Result := PAWN_INDEX;
        WHITE_ROOK:   Result := ROOK_INDEX;
        WHITE_KNIGHT: Result := KNIGHT_INDEX;
        WHITE_BISHOP: Result := BISHOP_INDEX;
        WHITE_QUEEN:  Result := QUEEN_INDEX;
        WHITE_KING:   Result := KING_INDEX;

        BLACK_PAWN:   Result := PAWN_INDEX;
        BLACK_ROOK:   Result := ROOK_INDEX;
        BLACK_KNIGHT: Result := KNIGHT_INDEX;
        BLACK_BISHOP: Result := BISHOP_INDEX;
        BLACK_QUEEN:  Result := QUEEN_INDEX;
        BLACK_KING:   Result := KING_INDEX;

        EMPTY_SQUARE: Result := EMPTY_INDEX;
    ELSE
        Result := UNKNOWN_INDEX;
    END;
END;

FUNCTION TChess.GetRevIndex(nBI: Integer): Integer;
BEGIN
    CASE (nBI) OF
        1: Result := 8;
        2: Result := 7;
        3: Result := 6;
        4: Result := 5;
        5: Result := 4;
        6: Result := 3;
        7: Result := 2;
        8: Result := 1;
    ELSE
        Result := nBI;
    END;
END;

FUNCTION TChess.IndexOk(nI: Integer): Boolean;
BEGIN
    Result := ((nI > 0) AND (nI < 9));
END;

FUNCTION TChess.GetPiece(nPc, nIdx: Integer): Char;
BEGIN
    CASE (nIdx) OF
        PAWN_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_PAWN
            ELSE
                Result := BLACK_PAWN;

        ROOK_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_ROOK
            ELSE
                Result := BLACK_ROOK;

        KNIGHT_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_KNIGHT
            ELSE
                Result := BLACK_KNIGHT;

        BISHOP_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_BISHOP
            ELSE
                Result := BLACK_BISHOP;

        QUEEN_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_QUEEN
            ELSE
                Result := BLACK_QUEEN;

        KING_INDEX:
            IF (nPc = WHITE_COLOR) THEN
                Result := WHITE_KING
            ELSE
                Result := BLACK_KING;
    ELSE
        Result := EMPTY_SQUARE;
    END;
END;

FUNCTION TChess.GetRevPiece(cPc: Char): Char;
BEGIN
    CASE (cPc) OF
        WHITE_PAWN:   Result := BLACK_PAWN;
        WHITE_ROOK:   Result := BLACK_ROOK;
        WHITE_KNIGHT: Result := BLACK_KNIGHT;
        WHITE_BISHOP: Result := BLACK_BISHOP;
        WHITE_QUEEN:  Result := BLACK_QUEEN;
        WHITE_KING:   Result := BLACK_KING;

        BLACK_PAWN:   Result := WHITE_PAWN;
        BLACK_ROOK:   Result := WHITE_ROOK;
        BLACK_KNIGHT: Result := WHITE_KNIGHT;
        BLACK_BISHOP: Result := WHITE_BISHOP;
        BLACK_QUEEN:  Result := WHITE_QUEEN;
        BLACK_KING:   Result := WHITE_KING;

        EMPTY_SQUARE: Result := EMPTY_SQUARE;
    ELSE
        Result := EMPTY_SQUARE;
    END;
END;

FUNCTION TChess.GetColor(cPc: Char): Integer;
BEGIN
    CASE (cPc) OF
        WHITE_PAWN:   Result := WHITE_COLOR;
        WHITE_ROOK:   Result := WHITE_COLOR;
        WHITE_KNIGHT: Result := WHITE_COLOR;
        WHITE_BISHOP: Result := WHITE_COLOR;
        WHITE_QUEEN:  Result := WHITE_COLOR;
        WHITE_KING:   Result := WHITE_COLOR;

        BLACK_PAWN:   Result := BLACK_COLOR;
        BLACK_ROOK:   Result := BLACK_COLOR;
        BLACK_KNIGHT: Result := BLACK_COLOR;
        BLACK_BISHOP: Result := BLACK_COLOR;
        BLACK_QUEEN:  Result := BLACK_COLOR;
        BLACK_KING:   Result := BLACK_COLOR;

        EMPTY_SQUARE: Result := EMPTY_COLOR;
    ELSE
        Result := UNKNOWN_COLOR;
    END;
END;

FUNCTION TChess.GetRevColor(cPc: Char): Integer;
BEGIN
    CASE (cPc) OF
        WHITE_PAWN:   Result := BLACK_COLOR;
        WHITE_ROOK:   Result := BLACK_COLOR;
        WHITE_KNIGHT: Result := BLACK_COLOR;
        WHITE_BISHOP: Result := BLACK_COLOR;
        WHITE_QUEEN:  Result := BLACK_COLOR;
        WHITE_KING:   Result := BLACK_COLOR;

        BLACK_PAWN:   Result := WHITE_COLOR;
        BLACK_ROOK:   Result := WHITE_COLOR;
        BLACK_KNIGHT: Result := WHITE_COLOR;
        BLACK_BISHOP: Result := WHITE_COLOR;
        BLACK_QUEEN:  Result := WHITE_COLOR;
        BLACK_KING:   Result := WHITE_COLOR;

        EMPTY_SQUARE: Result := EMPTY_COLOR;
    ELSE
        Result := UNKNOWN_COLOR;
    END;
END;

FUNCTION TChess.GetWhiteColor: Integer;
BEGIN
    Result := WHITE_COLOR;
END;

FUNCTION TChess.GetBlackColor: Integer;
BEGIN
    Result := BLACK_COLOR;
END;

FUNCTION TChess.GetWhitePawn: Char;
BEGIN
    Result := WHITE_PAWN;
END;

FUNCTION TChess.GetWhiteRook: Char;
BEGIN
    Result := WHITE_ROOK;
END;

FUNCTION TChess.GetWhiteKnight: Char;
BEGIN
    Result := WHITE_KNIGHT;
END;

FUNCTION TChess.GetWhiteBishop: Char;
BEGIN
    Result := WHITE_BISHOP;
END;

FUNCTION TChess.GetWhiteQueen: Char;
BEGIN
    Result := WHITE_QUEEN;
END;

FUNCTION TChess.GetWhiteKing: Char;
BEGIN
    Result := WHITE_KING;
END;

FUNCTION TChess.GetBlackPawn: Char;
BEGIN
    Result := BLACK_PAWN;
END;

FUNCTION TChess.GetBlackRook: Char;
BEGIN
    Result := BLACK_ROOK;
END;

FUNCTION TChess.GetBlackKnight: Char;
BEGIN
    Result := BLACK_KNIGHT;
END;

FUNCTION TChess.GetBlackBishop: Char;
BEGIN
    Result := BLACK_BISHOP;
END;

FUNCTION TChess.GetBlackQueen: Char;
BEGIN
    Result := BLACK_QUEEN;
END;

FUNCTION TChess.GetBlackKing: Char;
BEGIN
    Result := BLACK_KING;
END;

FUNCTION TChess.GetEmpty: Char;
BEGIN
    Result := EMPTY_SQUARE;
END;

FUNCTION TChess.GetPawnIndex: Integer;
BEGIN
    Result := PAWN_INDEX;
END;

FUNCTION TChess.GetRookIndex: Integer;
BEGIN
    Result := ROOK_INDEX;
END;

FUNCTION TChess.GetKnightIndex: Integer;
BEGIN
    Result := KNIGHT_INDEX;
END;

FUNCTION TChess.GetBishopIndex: Integer;
BEGIN
    Result := BISHOP_INDEX;
END;

FUNCTION TChess.GetQueenIndex: Integer;
BEGIN
    Result := QUEEN_INDEX;
END;

FUNCTION TChess.GetKingIndex: Integer;
BEGIN
    Result := KING_INDEX;
END;

FUNCTION TChess.IsWhite(cPc: Char): Boolean;
BEGIN
    Result := (GetColor(cPc) = WHITE_COLOR);
END;

FUNCTION TChess.IsBlack(cPc: Char): Boolean;
BEGIN
    Result := (GetColor(cPc) = BLACK_COLOR);
END;

FUNCTION TChess.IsPawn(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_PAWN) OR (cPc = BLACK_PAWN));
END;

FUNCTION TChess.IsRook(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_ROOK) OR (cPc = BLACK_ROOK));
END;

FUNCTION TChess.IsKnight(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_KNIGHT) OR (cPc = BLACK_KNIGHT));
END;

FUNCTION TChess.IsBishop(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_BISHOP) OR (cPc = BLACK_BISHOP));
END;

FUNCTION TChess.IsQueen(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_QUEEN) OR (cPc = BLACK_QUEEN));
END;

FUNCTION TChess.IsKing(cPc: Char): Boolean;
BEGIN
    Result := ((cPc = WHITE_KING) OR (cPc = BLACK_KING));
END;

FUNCTION TChess.IsWhitePawn(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_PAWN);
END;

FUNCTION TChess.IsWhiteRook(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_ROOK);
END;

FUNCTION TChess.IsWhiteKnight(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_KNIGHT);
END;

FUNCTION TChess.IsWhiteBishop(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_BISHOP);
END;

FUNCTION TChess.IsWhiteQueen(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_QUEEN);
END;

FUNCTION TChess.IsWhiteKing(cPc: Char): Boolean;
BEGIN
    Result := (cPc = WHITE_KING);
END;

FUNCTION TChess.IsBlackPawn(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_PAWN);
END;

FUNCTION TChess.IsBlackRook(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_ROOK);
END;

FUNCTION TChess.IsBlackKnight(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_KNIGHT);
END;

FUNCTION TChess.IsBlackBishop(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_BISHOP);
END;

FUNCTION TChess.IsBlackQueen(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_QUEEN);
END;

FUNCTION TChess.IsBlackKing(cPc: Char): Boolean;
BEGIN
    Result := (cPc = BLACK_KING);
END;

FUNCTION TChess.IsEmpty(cPc: Char): Boolean;
BEGIN
  Result:=(cPc = EMPTY_SQUARE);
END;

FUNCTION TChess.IsPromoteRank(nPc, nY: Integer): Boolean;
BEGIN
    Result := (nY = aPromoteY[nPc]);
END;

PROCEDURE TChess.AddMove(nFromX, nFromY, nToX, nToY: Integer;
                         cPiece, cPromote, cType: Char);
BEGIN
    IF (nMove < MAX_MOVE) THEN
    BEGIN
        nMove := nMove + 1;

        aMove[nMove].nFromX                   := nFromX;
        aMove[nMove].nFromY                   := nFromY;
        aMove[nMove].nToX                     := nToX;
        aMove[nMove].nToY                     := nToY;
        aMove[nMove].cPiece                   := cPiece;
        aMove[nMove].cPromote                 := cPromote;
        aMove[nMove].cCapture                 := aBoard[nToX, nToY];
        aMove[nMove].cType                    := cType;
        aMove[nMove].bCapturedIsPromotedPiece := aPromote[nToX, nToY];
    END;
END;

PROCEDURE TChess.GenPawnMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
BEGIN
    IF (NOT IndexOk(nFromY + aGenPawnAdvY[nPc])) THEN
        EXIT;

    IF (nFromY = aGenPawnOrgY[nPc]) THEN
    BEGIN
        IF (IsEmpty(aBoard[nFromX, aGenPawnOrgY[nPc] + aGenPawnAdvY[nPc]])) THEN
        BEGIN
            IF (IsEmpty(aBoard[nFromX, aGenPawnOrgY[nPc] +
                        aGenPawnAdvY[nPc] + aGenPawnAdvY[nPc]])) THEN
            BEGIN
                AddMove(nFromX, nFromY,
                        nFromX, aGenPawnOrgY[nPc] + aGenPawnAdvY[nPc] + aGenPawnAdvY[nPc],
                        cPc, EMPTY_SQUARE, MOVE_REGULAR);
            END;
        END;
    END;

    IF (nFromY = aPromoteY[nPc] - aGenPawnAdvY[nPc]) THEN
    BEGIN
        IF (IsEmpty(aBoard[nFromX, aPromoteY[nPc]])) THEN
        BEGIN
            AddMove(nFromX, nFromY,
                    nFromX, aPromoteY[nPc],
                    cPc, GetPiece(nPc, ROOK_INDEX), MOVE_PROMOTE);

            AddMove(nFromX, nFromY,
                    nFromX, aPromoteY[nPc],
                    cPc, GetPiece(nPc, KNIGHT_INDEX), MOVE_PROMOTE);

            AddMove(nFromX, nFromY,
                    nFromX, aPromoteY[nPc],
                    cPc, GetPiece(nPc, BISHOP_INDEX), MOVE_PROMOTE);

            AddMove(nFromX, nFromY,
                    nFromX, aPromoteY[nPc],
                    cPc, GetPiece(nPc, QUEEN_INDEX), MOVE_PROMOTE);
        END;

        IF (IndexOk(nFromX - 1)) THEN
        BEGIN
            IF (GetColor(aBoard[nFromX - 1, aPromoteY[nPc]]) = nRevPc) THEN
            BEGIN
                AddMove(nFromX, nFromY,
                        nFromX - 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, ROOK_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX - 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, KNIGHT_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX - 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, BISHOP_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX - 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, QUEEN_INDEX), MOVE_PROMOTE);
            END;
        END;

        IF (IndexOk(nFromX + 1)) THEN
        BEGIN
            IF (GetColor(aBoard[nFromX + 1, aPromoteY[nPc]]) = nRevPc) THEN
            BEGIN
                AddMove(nFromX, nFromY,
                        nFromX + 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, ROOK_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX + 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, KNIGHT_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX + 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, BISHOP_INDEX), MOVE_PROMOTE);

                AddMove(nFromX, nFromY,
                        nFromX + 1, aPromoteY[nPc],
                        cPc, GetPiece(nPc, QUEEN_INDEX), MOVE_PROMOTE);
            END;
        END;
    END
    ELSE
    BEGIN
        IF (IsEmpty(aBoard[nFromX, nFromY + aGenPawnAdvY[nPc]])) THEN
        BEGIN
            AddMove(nFromX, nFromY,
                    nFromX, nFromY + aGenPawnAdvY[nPc],
                    cPc, EMPTY_SQUARE, MOVE_REGULAR);
        END;

        IF (IndexOk(nFromX - 1)) THEN
        BEGIN
            IF (GetColor(aBoard[nFromX - 1, nFromY + aGenPawnAdvY[nPc]]) = nRevPc) THEN
            BEGIN
                AddMove(nFromX, nFromY,
                        nFromX - 1, nFromY + aGenPawnAdvY[nPc],
                        cPc, EMPTY_SQUARE, MOVE_REGULAR);
            END;
        END;

        IF (IndexOk(nFromX + 1)) THEN
        BEGIN
            IF (GetColor(aBoard[nFromX + 1, nFromY + aGenPawnAdvY[nPc]]) = nRevPc) THEN
            BEGIN
                AddMove(nFromX, nFromY,
                        nFromX + 1, nFromY + aGenPawnAdvY[nPc],
                        cPc, EMPTY_SQUARE, MOVE_REGULAR);
            END;
        END;
    END;

    IF (nFromY <> aGenPawnEPY[nPc]) THEN
        EXIT;
    IF (GetColor(mPrev.cPiece) <> nRevPc)  THEN
        EXIT;

    IF (GetIndex(mPrev.cPiece) <> PAWN_INDEX)  THEN
        EXIT;

    IF (mPrev.nFromY <> aGenPawnOrgY[nRevPc]) THEN
        EXIT;

    IF (mPrev.nToY <> aGenPawnEPY[nPc]) THEN
        EXIT;

    IF (mPrev.nToX = nFromX - 1) THEN
    BEGIN
        IF (IndexOk(nFromX - 1)) THEN
        BEGIN
            AddMove(nFromX, nFromY,
                    nFromX - 1, nFromY + aGenPawnAdvY[nPc],
                    cPc, EMPTY_SQUARE, MOVE_ENPASANT);
        END;
    END;

    IF (mPrev.nToX = nFromX + 1) THEN
    BEGIN
        IF (IndexOk(nFromX + 1)) THEN
        BEGIN
            AddMove(nFromX, nFromY,
                    nFromX + 1, nFromY + aGenPawnAdvY[nPc],
                    cPc, EMPTY_SQUARE, MOVE_ENPASANT);
        END;
    END;
END;

PROCEDURE TChess.GenCrossMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
VAR
    nI, nJ, nToX, nToY: Integer;
BEGIN
    FOR nI := 1 TO MAX_DIR DO
    BEGIN
        FOR nJ := 1 TO MAX_DIST DO
        BEGIN
            nToX := nFromX + aCrossX[nI, nJ];

            IF (NOT IndexOk(nToX)) THEN
                CONTINUE;

            nToY := nFromY + aCrossY[nI, nJ];

            IF (NOT IndexOk(nToY)) THEN
                CONTINUE;

            IF (IsEmpty(aBoard[nToX, nToY])) THEN
            BEGIN
                AddMove(nFromX, nFromY, nToX, nToY,
                        cPc, EMPTY_SQUARE, MOVE_REGULAR);
            END
            ELSE
            BEGIN
                IF (GetColor(aBoard[nToX, nToY]) = nRevPc) THEN
                BEGIN
                    AddMove(nFromX, nFromY, nToX, nToY,
                            cPc, EMPTY_SQUARE, MOVE_REGULAR);
                END;

                BREAK;
            END;
        END;
    END;
END;

PROCEDURE TChess.GenXShapeMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
VAR
    nI, nJ, nToX, nToY: Integer;
BEGIN
    FOR nI := 1 TO MAX_DIR DO
    BEGIN
        FOR nJ := 1 TO MAX_DIST DO
        BEGIN
            nToX := nFromX + aXShapeX[nI, nJ];

            IF (NOT IndexOk(nToX)) THEN
                CONTINUE;

            nToY := nFromY + aXShapeY[nI, nJ];

            IF (NOT IndexOk(nToY)) THEN
                CONTINUE;

            IF (IsEmpty(aBoard[nToX, nToY])) THEN
            BEGIN
                AddMove(nFromX, nFromY, nToX, nToY,
                        cPc, EMPTY_SQUARE, MOVE_REGULAR);
            END
            ELSE
            BEGIN
                IF (GetColor(aBoard[nToX, nToY]) = nRevPc) THEN
                BEGIN
                    AddMove(nFromX, nFromY, nToX, nToY,
                            cPc, EMPTY_SQUARE, MOVE_REGULAR);
                END;

                BREAK;
            END;
        END;
    END;
END;

PROCEDURE TChess.GenKnightMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
VAR
    nI, nToX, nToY: Integer;
BEGIN
    FOR nI := 1 TO MAX_KNIGHT DO
    BEGIN
        nToX := nFromX + aKnightX[nI];

        IF (NOT IndexOk(nToX)) THEN
            CONTINUE;

        nToY := nFromY + aKnightY[nI];

        IF (NOT IndexOk(nToY)) THEN
            CONTINUE;

        IF (IsEmpty(aBoard[nToX, nToY])) THEN
        BEGIN
            AddMove(nFromX, nFromY, nToX, nToY,
                    cPc, EMPTY_SQUARE, MOVE_REGULAR);
        END
        ELSE
        IF (GetColor(aBoard[nToX, nToY]) = nRevPc) THEN
        BEGIN
            AddMove(nFromX, nFromY, nToX, nToY,
                    cPc, EMPTY_SQUARE, MOVE_REGULAR);
        END;
    END;
END;

PROCEDURE TChess.GenKingMove(cPc: Char; nFromX, nFromY, nPc, nRevPc: Integer);
VAR
    nI, nToX, nToY: Integer;
BEGIN
    FOR nI := 1 TO MAX_KING DO
    BEGIN
        nToX := nFromX + aKingX[nI];

        IF (NOT IndexOk(nToX)) THEN
            CONTINUE;

        nToY := nFromY + aKingY[nI];

        IF (NOT IndexOk(nToY)) THEN
            CONTINUE;

        IF (IsEmpty(aBoard[nToX, nToY])) THEN
        BEGIN
            AddMove(nFromX, nFromY, nToX, nToY,
                    cPc, EMPTY_SQUARE, MOVE_REGULAR);
        END
        ELSE
        IF (GetColor(aBoard[nToX, nToY]) = nRevPc) THEN
        BEGIN
            AddMove(nFromX, nFromY, nToX, nToY,
                    cPc, EMPTY_SQUARE, MOVE_REGULAR);
        END;
    END;

    IF (aKCastle[nPc]) THEN
    BEGIN
        IF (nFromX = 5)                         AND
           (nFromY = aOrgKingY[nPc])            AND
           (IsEmpty(aBoard[6, aOrgKingY[nPc]])) AND
           (IsEmpty(aBoard[7, aOrgKingY[nPc]])) AND
           (aBoard[8, aOrgKingY[nPc]] = GetPiece(nPc, ROOK_INDEX)) THEN
        BEGIN
            IF (NOT SquareAttacked(6, aOrgKingY[nPc], nPc)) THEN
            BEGIN
                AddMove(nFromX, nFromY, 7, aOrgKingY[nPc],
                        cPc, EMPTY_SQUARE, MOVE_O_O);
            END;
        END;
    END;

    IF (aQCastle[nPc]) THEN
    BEGIN
        IF (nFromX = 5)                         AND
           (nFromY = aOrgKingY[nPc])            AND
           (IsEmpty(aBoard[2, aOrgKingY[nPc]])) AND
           (IsEmpty(aBoard[3, aOrgKingY[nPc]])) AND
           (IsEmpty(aBoard[4, aOrgKingY[nPc]])) AND
           (aBoard[1, aOrgKingY[nPc]] = GetPiece(nPc, ROOK_INDEX)) THEN
        BEGIN
            IF (NOT SquareAttacked(4, aOrgKingY[nPc], nPc)) THEN
            BEGIN
                AddMove(nFromX, nFromY, 3, aOrgKingY[nPc],
                        cPc, EMPTY_SQUARE, MOVE_O_O_O);
            END;
        END;
    END;
END;

PROCEDURE TChess.GenPieceMove(bWM: Boolean);
VAR
    nBX, nBY: Integer;
BEGIN
    IF (bWM) THEN
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            FOR nBX := 1 TO 8 DO
            BEGIN
                CASE (aBoard[nBX, nBY]) OF
                    WHITE_PAWN:
                        GenPawnMove(WHITE_PAWN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_ROOK:
                        GenCrossMove(WHITE_ROOK, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_KNIGHT:
                        GenKnightMove(WHITE_KNIGHT, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_BISHOP:
                        GenXShapeMove(WHITE_BISHOP, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_QUEEN:
                        BEGIN
                            GenCrossMove (WHITE_QUEEN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                            GenXShapeMove(WHITE_QUEEN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                        END;

                    WHITE_KING:
                        GenKingMove(WHITE_KING, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                END;
            END;
        END;
    END
    ELSE
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            FOR nBX := 1 TO 8 DO
            BEGIN
                CASE (aBoard[nBX, nBY]) OF
                    BLACK_PAWN:
                        GenPawnMove(BLACK_PAWN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_ROOK:
                        GenCrossMove(BLACK_ROOK, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_KNIGHT:
                        GenKnightMove(BLACK_KNIGHT, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_BISHOP:
                        GenXShapeMove(BLACK_BISHOP, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_QUEEN:
                        BEGIN
                            GenCrossMove(BLACK_QUEEN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                            GenXShapeMove(BLACK_QUEEN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                        END;

                    BLACK_KING:
                        GenKingMove(BLACK_KING, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                END;
            END;
        END;
    END;
END;

PROCEDURE TChess.GenOnePieceMove(bWM: Boolean; cPc: Char);
VAR
    nBX, nBY: Integer;
BEGIN
    IF (bWM) THEN
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            FOR nBX := 1 TO 8 DO
            BEGIN
                IF (aBoard[nBx, nBY] <> cPc) THEN
                BEGIN
                    CONTINUE;
                END;

                CASE (cPc) OF
                    WHITE_PAWN:
                        GenPawnMove(WHITE_PAWN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_ROOK:
                        GenCrossMove(WHITE_ROOK, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_KNIGHT:
                        GenKnightMove(WHITE_KNIGHT, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_BISHOP:
                        GenXShapeMove(WHITE_BISHOP, nBX, nBY, WHITE_COLOR, BLACK_COLOR);

                    WHITE_QUEEN:
                        BEGIN
                            GenCrossMove (WHITE_QUEEN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                            GenXShapeMove(WHITE_QUEEN, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                        END;

                    WHITE_KING:
                        GenKingMove(WHITE_KING, nBX, nBY, WHITE_COLOR, BLACK_COLOR);
                END;
            END;
        END;
    END
    ELSE
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            FOR nBX := 1 TO 8 DO
            BEGIN
                IF (aBoard[nBx, nBY] <> cPc) THEN
                BEGIN
                    CONTINUE;
                END;

                CASE (cPc) OF
                    BLACK_PAWN:
                        GenPawnMove(BLACK_PAWN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_ROOK:
                        GenCrossMove(BLACK_ROOK, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_KNIGHT:
                        GenKnightMove(BLACK_KNIGHT, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_BISHOP:
                        GenXShapeMove(BLACK_BISHOP, nBX, nBY, BLACK_COLOR, WHITE_COLOR);

                    BLACK_QUEEN:
                        BEGIN
                            GenCrossMove(BLACK_QUEEN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                            GenXShapeMove(BLACK_QUEEN, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                        END;

                    BLACK_KING:
                        GenKingMove(BLACK_KING, nBX, nBY, BLACK_COLOR, WHITE_COLOR);
                END;
            END;
        END;
    END;
END;

PROCEDURE TChess.GenDropMove(bWM: Boolean);
VAR
    nI, nBX, nBY: Integer;
BEGIN
    IF (bWM) THEN
    BEGIN
        FOR nI := 1 TO 5 DO
        BEGIN
            FOR nBY := 1 TO 8 DO
            BEGIN
                IF (aDropPiece[1, nI] = WHITE_PAWN) THEN
                BEGIN
                    IF (nBY = aPromoteY[WHITE_COLOR]) OR
                       (nBY = aPromoteY[BLACK_COLOR]) THEN
                        CONTINUE;
                END;

                FOR nBX := 1 TO 8 DO
                BEGIN
                    IF (NOT IsEmpty(aBoard[nBX, nBY])) THEN
                        CONTINUE;

                    AddMove(0, 0, nBX, nBY,
                            aDropPiece[1, nI], EMPTY_SQUARE, MOVE_DROP);
                END;
            END;
        END;
    END
    ELSE
    BEGIN
        FOR nI := 1 TO 5 DO
        BEGIN
            FOR nBY := 1 TO 8 DO
            BEGIN
                IF (aDropPiece[2, nI] = BLACK_PAWN) THEN
                BEGIN
                    IF (nBY = aPromoteY[WHITE_COLOR]) OR
                       (nBY = aPromoteY[BLACK_COLOR]) THEN
                        CONTINUE;
                END;

                FOR nBX := 1 TO 8 DO
                BEGIN
                    IF (NOT IsEmpty(aBoard[nBX, nBY])) THEN
                        CONTINUE;

                    AddMove(0, 0, nBX, nBY,
                            aDropPiece[2, nI], EMPTY_SQUARE, MOVE_DROP);
                END;
            END;
        END;
    END;
END;

PROCEDURE TChess.GenOneDropMove(bWM: Boolean; cPc: Char);
VAR
    nBX, nBY: Integer;
BEGIN
    IF (bWM) THEN
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            IF (cPc = WHITE_PAWN) THEN
            BEGIN
                IF (nBY = aPromoteY[WHITE_COLOR]) OR
                   (nBY = aPromoteY[BLACK_COLOR]) THEN
                    CONTINUE;
            END;

            FOR nBX := 1 TO 8 DO
            BEGIN
                IF (NOT IsEmpty(aBoard[nBX, nBY])) THEN
                    CONTINUE;

                AddMove(0, 0, nBX, nBY, cPc, EMPTY_SQUARE, MOVE_DROP);
            END;
        END;
    END
    ELSE
    BEGIN
        FOR nBY := 1 TO 8 DO
        BEGIN
            IF (cPc = BLACK_PAWN) THEN
            BEGIN
                IF (nBY = aPromoteY[WHITE_COLOR]) OR
                   (nBY = aPromoteY[BLACK_COLOR]) THEN
                    CONTINUE;
            END;

            FOR nBX := 1 TO 8 DO
            BEGIN
                IF (NOT IsEmpty(aBoard[nBX, nBY])) THEN
                    CONTINUE;

                AddMove(0, 0, nBX, nBY, cPc, EMPTY_SQUARE, MOVE_DROP);
            END;
        END;
    END;
END;

PROCEDURE TChess.GenMove;
BEGIN
    nMove := 0;
    GenPieceMove(bWhitesMove);
    GenDropMove (bWhitesMove);
END;

PROCEDURE TChess.BackupSetting;
VAR
    nI, nBY, nBX: Integer;
BEGIN
    FOR nI := 1 TO 2 DO
    BEGIN
        aTmpKingBX [nI] := aKingBX [nI];
        aTmpKingBY [nI] := aKingBY [nI];
        aTmpKCastle[nI] := aKCastle[nI];
        aTmpQCastle[nI] := aQCastle[nI];
    END;

    mTmpPrev.nFromX                   := mPrev.nFromX;
    mTmpPrev.nFromY                   := mPrev.nFromY;
    mTmpPrev.nToX                     := mPrev.nToX;
    mTmpPrev.nToY                     := mPrev.nToY;
    mTmpPrev.cPiece                   := mPrev.cPiece;
    mTmpPrev.cPromote                 := mPrev.cPromote;
    mTmpPrev.cCapture                 := mPrev.cCapture;
    mTmpPrev.cType                    := mPrev.cType;
    mTmpPrev.bCapturedIsPromotedPiece := mPrev.bCapturedIsPromotedPiece;

    FOR nBY := 1 TO 8 DO
    BEGIN
        FOR nBX := 1 TO 8 DO
        BEGIN
            aTmpBoard  [nBX, nBY] := aBoard  [nBX, nBY];
            aTmpPromote[nBX, nBY] := aPromote[nBX, nBY];
        END;
    END;
    
    FOR nBY := 1 TO 2 DO
    BEGIN
        FOR nBX := 1 TO 8 DO
        BEGIN
            aTmpcPiece[nBY, nBX] := acPiece[nBY, nBX];
            aTmpnPiece[nBY, nBX] := anPiece[nBY, nBX];
        END;
    END;

    bTmpWhitesMove := bWhitesMove;
END;

PROCEDURE TChess.RestoreSetting;
VAR
    nI, nBY, nBX: Integer;
BEGIN
    FOR nI := 1 TO 2 DO
    BEGIN
        aKingBX [nI] := aTmpKingBX [nI];
        aKingBY [nI] := aTmpKingBY [nI];
        aKCastle[nI] := aTmpKCastle[nI];
        aQCastle[nI] := aTmpQCastle[nI];
    END;

    mPrev.nFromX                   := mTmpPrev.nFromX;
    mPrev.nFromY                   := mTmpPrev.nFromY;
    mPrev.nToX                     := mTmpPrev.nToX;
    mPrev.nToY                     := mTmpPrev.nToY;
    mPrev.cPiece                   := mTmpPrev.cPiece;
    mPrev.cPromote                 := mTmpPrev.cPromote;
    mPrev.cCapture                 := mTmpPrev.cCapture;
    mPrev.cType                    := mTmpPrev.cType;
    mPrev.bCapturedIsPromotedPiece := mTmpPrev.bCapturedIsPromotedPiece;

    FOR nBY := 1 TO 8 DO
    BEGIN
        FOR nBX := 1 TO 8 DO
        BEGIN
            aBoard  [nBX, nBY] := aTmpBoard  [nBX, nBY];
            aPromote[nBX, nBY] := aTmpPromote[nBX, nBY];
        END;
    END;

    FOR nBY := 1 TO 2 DO
    BEGIN
        FOR nBX := 1 TO 8 DO
        BEGIN
            acPiece[nBY, nBX] := aTmpcPiece[nBY, nBX];
            anPiece[nBY, nBX] := aTmpnPiece[nBY, nBX];
        END;
    END;

    bWhitesMove := bTmpWhitesMove;
END;

FUNCTION TChess.SquareAttacked(nFromX, nFromY, nPc: Integer): Boolean;
VAR
    nRevPc, nI, nJ, nToX, nToY: Integer;
    bResult: Boolean;
BEGIN
    { Determine reverse color }
    IF (nPc = WHITE_COLOR) THEN
        nRevPc := BLACK_COLOR
    ELSE
        nRevPc := WHITE_COLOR;

    { Pawn }
    IF (IndexOk(nFromY + aGenPawnAdvY[nPc])) THEN
    BEGIN
        IF (IndexOk(nFromX - 1)) THEN
        BEGIN
            IF (aBoard[nFromX - 1, nFromY + aGenPawnAdvY[nPc]] =
                GetPiece(nRevPc, PAWN_INDEX)) THEN
            BEGIN
                Result := TRUE;
                EXIT;
            END;
        END;

        IF (IndexOk(nFromX + 1)) THEN
        BEGIN
            IF (aBoard[nFromX + 1, nFromY + aGenPawnAdvY[nPc]] =
                GetPiece(nRevPc, PAWN_INDEX)) THEN
            BEGIN
                Result := TRUE;
                EXIT;
            END;
        END;
    END;

    { Rook & queen }
    bResult := FALSE;
    FOR nI  := 1 TO MAX_DIR DO
    BEGIN
        FOR nJ := 1 TO MAX_DIST DO
        BEGIN
            nToX := nFromX + aCrossX[nI, nJ];

            IF (NOT IndexOk(nToX)) THEN
                CONTINUE;

            nToY := nFromY + aCrossY[nI, nJ];

            IF (NOT IndexOk(nToY)) THEN
                CONTINUE;

            IF (IsEmpty(aBoard[nToX, nToY])) THEN
                CONTINUE;

            bResult := (aBoard[nToX, nToY] = GetPiece(nRevPc, ROOK_INDEX )) OR
                       (aBoard[nToX, nToY] = GetPiece(nRevPc, QUEEN_INDEX));
            BREAK;
        END;
        IF (bResult) THEN
            BREAK;
    END;
    IF (bResult) THEN
    BEGIN
        Result := TRUE;
        EXIT;
    END;

    { Bishop }
    FOR nI := 1 TO MAX_DIR DO
    BEGIN
        FOR nJ := 1 TO MAX_DIST DO
        BEGIN
            nToX := nFromX + aXShapeX[nI, nJ];

            IF (NOT IndexOk(nToX)) THEN
                CONTINUE;

            nToY := nFromY + aXShapeY[nI, nJ];

            IF (NOT IndexOk(nToY)) THEN
                CONTINUE;

            IF (IsEmpty(aBoard[nToX, nToY])) THEN
                CONTINUE;

            bResult := (aBoard[nToX, nToY] = GetPiece(nRevPc, BISHOP_INDEX)) OR
                       (aBoard[nToX, nToY] = GetPiece(nRevPc, QUEEN_INDEX));
            BREAK;
        END;
        IF (bResult) THEN
            BREAK;
    END;
    IF (bResult) THEN
    BEGIN
        Result := TRUE;
        EXIT;
    END;

    { Knight }
    FOR nI := 1 TO MAX_KNIGHT DO
    BEGIN
        nToX := nFromX + aKnightX[nI];

        IF (NOT IndexOk(nToX)) THEN
            CONTINUE;

        nToY := nFromY + aKnightY[nI];

        IF (NOT IndexOk(nToY)) THEN
            CONTINUE;

        IF (IsEmpty(aBoard[nToX, nToY])) THEN
            CONTINUE;

        bResult := (aBoard[nToX, nToY] = GetPiece(nRevPc, KNIGHT_INDEX));

        IF (bResult) THEN
            BREAK;
    END;
    IF (bResult) THEN
    BEGIN
        Result := TRUE;
        EXIT;
    END;

    { King }
    FOR nI := 1 TO MAX_KING DO
    BEGIN
        nToX := nFromX + aKingX[nI];

        IF (NOT IndexOk(nToX)) THEN
            CONTINUE;

        nToY := nFromY + aKingY[nI];

        IF (NOT IndexOk(nToY)) THEN
            CONTINUE;

        IF (IsEmpty(aBoard[nToX, nToY])) THEN
            CONTINUE;

        bResult := (aBoard[nToX, nToY] = GetPiece(nRevPc, KING_INDEX));

        IF (bResult) THEN
            BREAK;
    END;

    Result := bResult;
END;

FUNCTION TChess.SmartMove(nFromX, nFromY: Integer): Integer;
VAR
    nPromX, nPromY, nResult, nI, nCnt, nPc: Integer;
BEGIN
    nPromX  := 0;
    nPromY  := 0;
    nResult := 0;
    nCnt    := 0;
    nPc     := GetColor(aBoard[nFromX, nFromY]);

    FOR nI := 1 TO nMove DO
    BEGIN
        IF (aMove[nI].nFromX <> nFromX) THEN
            CONTINUE;

        IF (aMove[nI].nFromY <> nFromY) THEN
            CONTINUE;

        BackupSetting;

        ExecMove(nI);

        IF (SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
        BEGIN
            RestoreSetting;
        END
        ELSE
        BEGIN
            RestoreSetting;

            IF (aMove[nI].cType = MOVE_PROMOTE) THEN
            BEGIN
                IF ((aMove[nI].nToX <> nPromX) OR (aMove[nI].nToY <> nPromY)) THEN
                BEGIN
                    IF (IsQueen(aMove[nI].cPromote)) THEN
                    BEGIN
                        nCnt   := nCnt + 1;
                        nPromX := aMove[nI].nToX;
                        nPromY := aMove[nI].nToY;

                        IF (nCnt > 1) THEN
                        BEGIN
                            nResult := 0;
                            BREAK;
                        END;

                        nResult := nI;
                    END;    
                END;
            END
            ELSE
            BEGIN
                nCnt := nCnt + 1;

                IF (nCnt > 1) THEN
                BEGIN
                    nResult := 0;
                    BREAK;
                END;

                nResult := nI;
            END;
        END;
    END;

    Result := nResult;
END;

FUNCTION TChess.MatchMove(nFromX, nFromY, nToX, nToY: Integer;
                          cPiece, cPromote: Char): Integer;
VAR
    nResult, nI, nPc: Integer;
BEGIN
    nResult := 0;

    FOR nI := 1 TO nMove DO
    BEGIN
        IF (aMove[nI].nFromX <> nFromX) THEN
            CONTINUE;

        IF (aMove[nI].nFromY <> nFromY) THEN
            CONTINUE;

        IF (aMove[nI].nToX <> nToX) THEN
            CONTINUE;

        IF (aMove[nI].nToY <> nToY) THEN
            CONTINUE;

        IF (aMove[nI].cPiece <> cPiece) THEN
            CONTINUE;

        IF (aMove[nI].cPromote <> cPromote) THEN
            CONTINUE;

        nResult := nI;
        BREAK;
    END;

    IF (nResult <> 0) THEN
    BEGIN
        BackupSetting;

        ExecMove(nResult);

        nPc := GetColor(cPiece);

        IF (SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
        BEGIN
            nResult := 0;
        END;
        
        RestoreSetting;
    END;

    Result := nResult;
END;

FUNCTION TChess.LegalMove(nFromX, nFromY, nToX, nToY: Integer;
                          cPiece, cPromote: Char): Integer;
BEGIN
    Result := 0;

    IF ((nFromX <> 0) OR (nFromY <> 0)) THEN
    BEGIN
        IF (NOT IndexOk(nFromX)) THEN
            EXIT;

        IF (NOT IndexOk(nFromY)) THEN
            EXIT;

        IF (cPiece <> aBoard[nFromX, nFromY]) THEN
            EXIT;
    END;

    IF (IsEmpty(cPiece)) THEN
        EXIT;

    IF (NOT IndexOk(nToX)) THEN
        EXIT;

    IF (NOT IndexOk(nToY)) THEN
        EXIT;

    IF (bWhitesMove) THEN
    BEGIN
        IF (GetColor(cPiece) <> WHITE_COLOR) THEN
            EXIT;
    END
    ELSE
    BEGIN
        IF (GetColor(cPiece) <> BLACK_COLOR) THEN
            EXIT;
    END;

    IF (GetColor(aBoard[nToX, nToY]) = GetColor(cPiece)) THEN
        EXIT;

    Result := MatchMove(nFromX, nFromY, nToX, nToY, cPiece, cPromote);
END;

FUNCTION TChess.ExecMove(nMI: Integer): Integer;
VAR
    nResult, nPc, nPi, nI: Integer;
BEGIN
    nResult := 0;

    nPc     := GetColor(aMove[nMI].cPiece);
    nPi     := GetIndex(aMove[nMI].cPiece);


    IF (aMove[nMI].cType = MOVE_DROP) THEN
    BEGIN
        aBoard  [aMove[nMI].nToX, aMove[nMI].nToY] := aMove[nMI].cPiece;
        aPromote[aMove[nMI].nToX, aMove[nMI].nToY] := FALSE;

        FOR nI := 1 TO 5 DO
        BEGIN
//            IF (acPiece[nPc, nI] = aMove[nMI].cPiece) THEN
//            if adroppiece[npc,ni]=amove[nmi].cPiece then
            if ((acPiece[nPc, nI] = aMove[nMI].cPiece) or ((bairdrop) and  (adroppiece[npc,ni]=amove[nmi].cPiece))) then
            BEGIN
//                IF (anPiece[nPc, nI] > 0) THEN
                    if ((bairdrop) or (anPiece[nPc, nI] > 0)) then
                    anPiece[nPc, nI] := anPiece[nPc, nI] - 1;

                IF (anPiece[nPc, nI] = 0) THEN
                BEGIN
                    acPiece[nPc, nI] := EMPTY_SQUARE;
                END
                else acpiece[npc,ni]:=adroppiece[npc,ni];

                BREAK;
            END;
        END;
    END
    ELSE
    BEGIN
        IF (aMove[nMI].cType = MOVE_PROMOTE) THEN
        BEGIN
            aBoard  [aMove[nMI].nToX, aMove[nMI].nToY] := aMove[nMI].cPromote;
            aPromote[aMove[nMI].nToX, aMove[nMI].nToY] := TRUE;
        END
        ELSE
        BEGIN
            aBoard  [aMove[nMI].nToX, aMove[nMI].nToY] := aMove[nMI].cPiece;
            aPromote[aMove[nMI].nToX, aMove[nMI].nToY] := aPromote[aMove[nMI].nFromX, aMove[nMI].nFromY];
        END;

        aBoard  [aMove[nMI].nFromX, aMove[nMI].nFromY] := EMPTY_SQUARE;
        aPromote[aMove[nMI].nFromX, aMove[nMI].nFromY] := FALSE;

        { Check enpasant }
        IF (aMove[nMI].cType = MOVE_ENPASANT) THEN
        BEGIN
            aBoard  [aMove[nMI].nToX, aMove[nMI].nFromY] := EMPTY_SQUARE;
            aPromote[aMove[nMI].nToX, aMove[nMI].nFromY] := FALSE;

            nResult := 1;
        END;

        { Check king side castling move }
        IF (aMove[nMI].cType = MOVE_O_O) THEN
        BEGIN
            aBoard  [8, aOrgKingY[nPc]] := EMPTY_SQUARE;
            aBoard  [6, aOrgKingY[nPc]] := GetPiece(nPc, ROOK_INDEX);
            aPromote[8, aOrgKingY[nPc]] := FALSE;
            aPromote[6, aOrgKingY[nPc]] := FALSE;

            nResult := 2;
        END;

        { Check queen side castling move }
        IF (aMove[nMI].cType = MOVE_O_O_O) THEN
        BEGIN
            aBoard  [1, aOrgKingY[nPc]] := EMPTY_SQUARE;
            aBoard  [4, aOrgKingY[nPc]] := GetPiece(nPc, ROOK_INDEX);
            aPromote[1, aOrgKingY[nPc]] := FALSE;
            aPromote[4, aOrgKingY[nPc]] := FALSE;

            nResult := 3;
        END;
    END;

    { Save current move into previous move record }
    mPrev.nFromX                   := aMove[nMI].nFromX;
    mPrev.nFromY                   := aMove[nMI].nFromY;
    mPrev.nToX                     := aMove[nMI].nToX;
    mPrev.nToY                     := aMove[nMI].nToY;
    mPrev.cPiece                   := aMove[nMI].cPiece;
    mPrev.cPromote                 := aMove[nMI].cPromote;
    mPrev.cCapture                 := aMove[nMI].cCapture;
    mPrev.cType                    := aMove[nMI].cType;
    mPrev.bCapturedIsPromotedPiece := aMove[nMI].bCapturedIsPromotedPiece;

    { Set castling flags }
    IF (nPi = KING_INDEX) THEN
    BEGIN
        aKCastle[nPc] := FALSE;
        aQCastle[nPc] := FALSE;
    END;

    IF (nPi = ROOK_INDEX) THEN
    BEGIN
        IF ((aMove[nMI].nFromX = 1) AND (aMove[nMI].nFromY = aOrgRookY[nPc])) THEN
            aQCastle[nPc] := FALSE;

        IF ((aMove[nMI].nFromX = 8) AND (aMove[nMI].nFromY = aOrgRookY[nPc])) THEN
            aKCastle[nPc] := FALSE;
    END;

    { Set king position }
    IF (nPi = KING_INDEX) THEN
    BEGIN
        aKingBX[nPc] := aMove[nMI].nToX;
        aKingBY[nPc] := aMove[nMI].nToY;
    END;

    { Flip move flag }
    bWhitesMove := (NOT bWhitesMove);

    { Return result }
    Result := nResult;
END;

FUNCTION TChess.GetBoardStr: String;
VAR
    cSt:    String;
    nY, nX: Integer;
BEGIN
    cSt := '';
    FOR nY := 1 TO 8 DO
        FOR nX := 1 TO 8 DO
            cSt := cSt + aBoard[nX, nY];
    Result := cSt;
END;

FUNCTION TChess.GetWhitePieceStr: String;
VAR
    cSt:    String;
    nI, nJ: Integer;
BEGIN
    cSt := '';
    FOR nI := 1 TO 5 DO
        FOR nJ := 1 TO anPiece[1, nI] DO
            cSt := cSt + acPiece[1, nI];
    Result := cSt;
END;

FUNCTION TChess.GetBlackPieceStr: String;
VAR
    cSt:    String;
    nI, nJ: Integer;
BEGIN
    cSt := '';
    FOR nI := 1 TO 5 DO
        FOR nJ := 1 TO anPiece[2, nI] DO
            cSt := cSt + acPiece[2, nI];
    Result := cSt;
END;

FUNCTION TChess.GetFENBoardStr: String;
VAR
    cSt, cTmp: String;
    nY, nX, nE: Integer;
BEGIN
    cSt := '';
    FOR nY := 8 DOWNTO 1 DO
    BEGIN
        cTmp := '';
        nE   := 0;
        FOR nX := 1 TO 8 DO
        BEGIN
            IF (IsEmpty(aBoard[nX, nY])) THEN
                Inc(nE)
            ELSE
            IF (nE > 0) THEN
            BEGIN
                cTmp := cTmp + IntToStr(nE) + aBoard[nX, nY];
                if aPromote[nX,nY] then ctmp:=ctmp+'~';
                nE   := 0;
            END
            ELSE begin
                cTmp := cTmp + aBoard[nX, nY];
                if aPromote[nX,nY] then ctmp:=ctmp+'~';
            end;
        END;

        IF (nE > 0) THEN
            cSt := cSt + cTmp + IntToStr(nE)
        ELSE
            cSt := cSt + cTmp;

        IF (nY > 1) THEN
            cSt := cSt + '/';
    END;
    cst:=cst+'/'+getwhitePieceStr+getBlackPieceStr;

    IF (bWhitesMove) THEN
        cSt := cSt + ' w '
    ELSE
        cSt := cSt + ' b ';

    IF ((NOT aKCastle[WHITE_COLOR]) AND (NOT aQCastle[WHITE_COLOR]) AND
        (NOT aKCastle[BLACK_COLOR]) AND (NOT aQCastle[BLACK_COLOR])) THEN
        cSt := cSt + '-'
    ELSE
    BEGIN
        IF (aKCastle[WHITE_COLOR]) THEN cSt := cSt + 'K';
        IF (aQCastle[WHITE_COLOR]) THEN cSt := cSt + 'Q';
        IF (aKCastle[BLACK_COLOR]) THEN cSt := cSt + 'k';
        IF (aQCastle[BLACK_COLOR]) THEN cSt := cSt + 'q';
    END;
//    cSt := cSt + ' - 0 1';
    cst:=cst+' '+getenpassantsq;
    Result := cSt;
END;

FUNCTION TChess.GetMoveStr(nMI: Integer; bCheckKing, bMateKing: Boolean): String;
VAR
    cSt: String;
    nI, nPc, nX, nY: Integer;
    bResult: Boolean;
BEGIN
    cSt := UpperCase(aMove[nMI].cPiece);

    IF (aMove[nMI].cType = MOVE_PROMOTE) THEN
    BEGIN
        IF (IsEmpty(aMove[nMI].cCapture)) THEN
            cSt := Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY) +
                   '=' + UpperCase(aMove[nMI].cPromote)
        ELSE
            cSt := Chr(96 + aMove[nMI].nFromX) + 'x' + Chr(96 + aMove[nMI].nToX) +
                   IntToStr(aMove[nMI].nToY) +
                   '=' + UpperCase(aMove[nMI].cPromote);
    END
    ELSE
    BEGIN
        IF (aMove[nMI].cType = MOVE_DROP) THEN
        BEGIN
            { Set piece drop move string }
            cSt := cSt + '@' + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
        END
        ELSE
        BEGIN
            IF (IsEmpty(aMove[nMI].cCapture)) THEN
            BEGIN
                IF (GetIndex(aMove[nMI].cPiece) = PAWN_INDEX) THEN
                    cSt := Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY)
                ELSE
                BEGIN
                    nX      := 0;
                    nY      := 0;
                    bResult := FALSE;
                    FOR nI := 1 TO nMove DO
                    BEGIN
                        IF ((aMove[nI].nFromX = aMove[nMI].nFromX) AND
                            (aMove[nI].nFromY = aMove[nMI].nFromY)) THEN
                            CONTINUE;

                        IF (aMove[nI].nToX <> aMove[nMI].nToX) THEN
                            CONTINUE;

                        IF (aMove[nI].nToY <> aMove[nMI].nToY) THEN
                            CONTINUE;

                        IF (aMove[nI].cPiece <> aMove[nMI].cPiece) THEN
                            CONTINUE;

                        IF (aMove[nI].cType <> aMove[nMI].cType) THEN
                            CONTINUE;

                        bResult := TRUE;

                        IF (aMove[nI].nFromX = aMove[nMI].nFromX) THEN
                            Inc(nX);

                        IF (aMove[nI].nFromY = aMove[nMI].nFromY) THEN
                            Inc(nY);
                    END;

                    IF ((nX > 0) AND (nY > 0)) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) + IntToStr(aMove[nMI].nFromY) +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY  )
                    ELSE
                    IF (nY > 0) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                    IF (nX > 0) THEN
                        cSt := cSt + IntToStr(aMove[nMI].nFromY) +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                    IF (bResult) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                        cSt := cSt + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY)
                END;
            END
            ELSE
            BEGIN
                IF (GetIndex(aMove[nMI].cPiece) = PAWN_INDEX) THEN
                    cSt := Chr(96 + aMove[nMI].nFromX) + 'x' +
                           Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY)
                ELSE
                BEGIN
                    nX      := 0;
                    nY      := 0;
                    bResult := FALSE;
                    FOR nI  := 1 TO nMove DO
                    BEGIN
                        IF ((aMove[nI].nFromX = aMove[nMI].nFromX) AND
                            (aMove[nI].nFromY = aMove[nMI].nFromY)) THEN
                            CONTINUE;

                        IF (aMove[nI].nToX <> aMove[nMI].nToX) THEN
                            CONTINUE;

                        IF (aMove[nI].nToY <> aMove[nMI].nToY) THEN
                            CONTINUE;

                        IF (aMove[nI].cPiece <> aMove[nMI].cPiece) THEN
                            CONTINUE;

                        IF (aMove[nI].cType <> aMove[nMI].cType) THEN
                            CONTINUE;

                        bResult := TRUE;

                        IF (aMove[nI].nFromX = aMove[nMI].nFromX) THEN
                            Inc(nX);

                        IF (aMove[nI].nFromY = aMove[nMI].nFromY) THEN
                            Inc(nY);
                    END;

                    IF ((nX > 0) AND (nY > 0)) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) + IntToStr(aMove[nMI].nFromY) + 'x' +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY  )
                    ELSE
                    IF (nY > 0) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) + 'x' +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                    IF (nX > 0) THEN
                        cSt := cSt + IntToStr(aMove[nMI].nFromY) + 'x' +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                    IF (bResult) THEN
                        cSt := cSt + Chr(96 + aMove[nMI].nFromX) + 'x' +
                                     Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY)
                    ELSE
                        cSt := cSt + 'x' + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
                END;
            END;
        END;
    END;

    { Check enpasant }
    IF (aMove[nMI].cType = MOVE_ENPASANT) THEN
    BEGIN
        { Set move string }
        cSt := Chr(96 + aMove[nMI].nFromX) + 'x' +
               Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
    END;

    { Check king side castling move }
    IF (aMove[nMI].cType = MOVE_O_O) THEN
    BEGIN
        { Set move string }
        cSt := 'O-O';
    END;

    { Check queen side castling move }
    IF (aMove[nMI].cType = MOVE_O_O_O) THEN
    BEGIN
        { Set move string }
        cSt := 'O-O-O';
    END;

    { If no need to check king }
    IF (NOT bCheckKing) THEN
    BEGIN
        Result := cSt;
        EXIT;
    END;

    { Get reverse piece color }
    nPc := GetRevColor(aMove[nMI].cPiece);

    { Check to see if opponent's king in check }
    IF (NOT IndexOk(aKingBX[nPc])) THEN
    BEGIN
        Result := cSt;
        EXIT;
    END;

    IF (NOT IndexOk(aKingBY[nPc])) THEN
    BEGIN
        Result := cSt;
        EXIT;
    END;

    IF (NOT SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
    BEGIN
        Result := cSt;
        EXIT;
    END;

    IF (NOT bMateKing) THEN
    BEGIN
        Result := cSt + '+';
        EXIT;
    END;

    { In check, check to see if opponent's king is mated }
    GenMove;

    bResult := TRUE;
    
    FOR nI := 1 TO nMove DO
    BEGIN
        BackupSetting;

        ExecMove(nI);

        IF (SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
        BEGIN
            RestoreSetting;
            CONTINUE;
        END;

        RestoreSetting;
        
        bResult := FALSE;
        BREAK;
    END;

    IF (bResult) THEN
        Result := cSt + '#'
    ELSE
        Result := cSt + '+';
END;

PROCEDURE TChess.GetAlgebraicMoveStr(nMI: Integer; bCheckKing, bMateKing: Boolean; VAR cM1, cM2, cM3, cM4: String);
VAR
    cSt: String;
    nI, nPc: Integer;
    bResult: Boolean;
BEGIN
    cM1 := '';
    cM2 := '';
    cM3 := '';
    cM4 := '';
    cSt := UpperCase(aMove[nMI].cPiece);

    IF (aMove[nMI].cType = MOVE_PROMOTE) THEN
    BEGIN
        IF (IsEmpty(aMove[nMI].cCapture)) THEN
            cSt := Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY) +
                   '=' + UpperCase(aMove[nMI].cPromote)
        ELSE
            cSt := Chr(96 + aMove[nMI].nFromX) + 'x' + Chr(96 + aMove[nMI].nToX) +
                   IntToStr(aMove[nMI].nToY) +
                   '=' + UpperCase(aMove[nMI].cPromote);

        cM1 := cSt;
        cM2 := cSt;
        cM3 := cSt;
        cM4 := cSt;
    END
    ELSE
    BEGIN
        IF (aMove[nMI].cType = MOVE_DROP) THEN
        BEGIN
            { Set piece drop move string }
            cSt := cSt + '@' + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
            cM1 := cSt;
            cM2 := cSt;
            cM3 := cSt;
            cM4 := cSt;
        END
        ELSE
        BEGIN
            IF (IsEmpty(aMove[nMI].cCapture)) THEN
            BEGIN
                IF (GetIndex(aMove[nMI].cPiece) = PAWN_INDEX) THEN
                BEGIN
                    cSt := Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
                    cM1 := cSt;
                    cM2 := cSt;
                    cM3 := cSt;
                    cM4 := cSt;
                END
                ELSE
                BEGIN
                    cM1 := cSt + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);

                    cM2 := cSt + Chr(96 + aMove[nMI].nFromX) +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY);

                    cM3 := cSt + IntToStr(aMove[nMI].nFromY) +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY);

                    cM4 := cSt + Chr(96 + aMove[nMI].nFromX) + IntToStr(aMove[nMI].nFromY) +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY  );
                END;
            END
            ELSE
            BEGIN
                IF (GetIndex(aMove[nMI].cPiece) = PAWN_INDEX) THEN
                BEGIN
                    cSt := Chr(96 + aMove[nMI].nFromX) + 'x' +
                           Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
                    cM1 := cSt;
                    cM2 := cSt;
                    cM3 := cSt;
                    cM4 := cSt;
                END
                ELSE
                BEGIN
                    cM1 := cSt + 'x' + Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);

                    cM2 := cSt + Chr(96 + aMove[nMI].nFromX) + 'x' +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY);

                    cM3 := cSt + IntToStr(aMove[nMI].nFromY) + 'x' +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY);

                    cM4 := cSt + Chr(96 + aMove[nMI].nFromX) + IntToStr(aMove[nMI].nFromY) + 'x' +
                                 Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY  );
                END;
            END;
        END;
    END;

    { Check enpasant }
    IF (aMove[nMI].cType = MOVE_ENPASANT) THEN
    BEGIN
        { Set move string }
        cSt := Chr(96 + aMove[nMI].nFromX) + 'x' +
               Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
        cM1 := cSt;
        cM2 := cSt;
        cM3 := cSt;
        cM4 := cSt;
    END;

    { Check king side castling move }
    IF (aMove[nMI].cType = MOVE_O_O) THEN
    BEGIN
        { Set move string }
        cSt := 'O-O';
        cM1 := cSt;
        cM2 := cSt;
        cM3 := cSt;
        cM4 := cSt;
    END;

    { Check queen side castling move }
    IF (aMove[nMI].cType = MOVE_O_O_O) THEN
    BEGIN
        { Set move string }
        cSt := 'O-O-O';
        cM1 := cSt;
        cM2 := cSt;
        cM3 := cSt;
        cM4 := cSt;
    END;

    { If no need to check king }
    IF (NOT bCheckKing) THEN
        EXIT;

    { Get reverse piece color }
    nPc := GetRevColor(aMove[nMI].cPiece);

    { Check to see if opponent's king in check }
    IF (NOT IndexOk(aKingBX[nPc])) THEN
        EXIT;

    IF (NOT IndexOk(aKingBY[nPc])) THEN
        EXIT;

    IF (NOT SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
        EXIT;

    IF (NOT bMateKing) THEN
    BEGIN
        cM1 := cM1 + '+';
        cM2 := cM2 + '+';
        cM3 := cM3 + '+';
        cM4 := cM4 + '+';
        EXIT;
    END;

    { In check, check to see if opponent's king is mated }
    GenMove;

    bResult := TRUE;
    
    FOR nI := 1 TO nMove DO
    BEGIN
        BackupSetting;

        ExecMove(nI);

        IF (SquareAttacked(aKingBX[nPc], aKingBY[nPc], nPc)) THEN
        BEGIN
            RestoreSetting;
            CONTINUE;
        END;

        RestoreSetting;
        
        bResult := FALSE;
        BREAK;
    END;

    IF (bResult) THEN
    BEGIN
        cM1 := cM1 + '#';
        cM2 := cM2 + '#';
        cM3 := cM3 + '#';
        cM4 := cM4 + '#';
    END
    ELSE
    BEGIN
        cM1 := cM1 + '+';
        cM2 := cM2 + '+';
        cM3 := cM3 + '+';
        cM4 := cM4 + '+';
    END;
END;

FUNCTION TChess.GetLongMoveStr(nMI: Integer): String;
BEGIN
    IF (aMove[nMI].cType = MOVE_DROP) THEN
    BEGIN
        Result := UpperCase(aMove[nMI].cPiece) + '@' +
                  Chr(96 + aMove[nMI].nToX) + IntToStr(aMove[nMI].nToY);
    END
    ELSE
    BEGIN
        Result := Chr(96 + aMove[nMI].nFromX) + IntToStr(aMove[nMI].nFromY) +
                  Chr(96 + aMove[nMI].nToX  ) + IntToStr(aMove[nMI].nToY  );
    END;
END;

FUNCTION TChess.ExecAlgebraicMove(nMoveNum: Integer; cBoard: Char; cOrgMove: String; VAR nMI: Integer): String;
VAR
    cMove, cM1, cM2, cM3, cM4: String;
    nS, nI, nJ, nC: Integer;
BEGIN
    nMI := 0;

    { Check size }
    cMove := Trim(cOrgMove);
    nS    := Length(cMove);
    IF (nS = 0) THEN
    BEGIN
        Result := 'Missing move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
        EXIT;
    END;

    { Remove + and # }
    nJ := 0;
    FOR nI := nS DOWNTO 1 DO
    BEGIN
        IF ((cMove[nI] <> '+') AND (cMove[nI] <> '#')) THEN
        BEGIN
            nJ := nI;
            BREAK;
        END;
    END;
    IF (nJ = 0) THEN
    BEGIN
        Result := 'Missing move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
        EXIT;
    END;
    cMove := Copy(cMove, 1, nJ);

    { Check size }
    nS := Length(cMove);
    IF (nS = 0) THEN
    BEGIN
        Result := 'Missing move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
        EXIT;
    END;

    { Reset possible moves }
    nMove := 0;

    { Get color move index }
    IF (bWhitesMove) THEN
        nC := WHITE_COLOR
    ELSE
        nC := BLACK_COLOR;

    { Drop move? }
    IF (cMove[2] = '@') THEN
    BEGIN
        IF (cMove[1] = 'P') THEN
        BEGIN
            if ((acpiece[nc,pawn_index]=EMPTY_SQUARE) and (not bairdrop)) then
            begin
             Result:= 'Illegal drop [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
             exit;
            end;
            GenOneDropMove(bWhitesMove, aDropPiece[nC, pawn_index]);
        END
        ELSE
        IF (cMove[1] = 'R') THEN
        BEGIN
            if ((acpiece[nc,rook_index]=EMPTY_SQUARE) and (not bairdrop)) then
            begin
             Result:= 'Illegal drop [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
             exit;
            end;
            GenOneDropMove(bWhitesMove, aDropPiece[nC, rook_index]);
        END
        ELSE
        IF (cMove[1] = 'N') THEN
        BEGIN
            if ((acpiece[nc,knight_index]=EMPTY_SQUARE) and (not bairdrop)) then
            begin
             Result:= 'Illegal drop [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
             exit;
            end;
            GenOneDropMove(bWhitesMove, aDropPiece[nC, knight_index]);
        END
        ELSE
        IF (cMove[1] = 'B') THEN
        BEGIN
            if ((acpiece[nc,bishop_index]=EMPTY_SQUARE) and (not bairdrop)) then
            begin
             Result:= 'Illegal drop [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
             exit;
            end;
            GenOneDropMove(bWhitesMove, aDropPiece[nC, bishop_index]);
        END
        ELSE
        IF (cMove[1] = 'Q') THEN
        BEGIN
            if ((acpiece[nc,queen_index]=EMPTY_SQUARE) and (not bairdrop)) then
            begin
             Result:= 'Illegal drop [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + ']';
             exit;
            end;
            GenOneDropMove(bWhitesMove, aDropPiece[nC, queen_index]);
        END
        ELSE
        BEGIN
            Result := 'Unknown piece drop move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + '].';
            EXIT;
        END;
    END;

    { Piece move? }
    IF ((cMove[1] = 'P') OR (cMove[1] = 'a') OR (cMove[1] = 'c') OR
        (cMove[1] = 'd') OR (cMove[1] = 'e') OR (cMove[1] = 'f') OR
        (cMove[1] = 'g') OR (cMove[1] = 'h')) THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 1]);
    END
    ELSE
    IF (cMove[1] = 'b') THEN
    BEGIN
        IF (bWhitesMove) THEN
            GenOnePieceMove(bWhitesMove, aDropPiece[nC, 1])
        ELSE
        BEGIN
            GenOnePieceMove(bWhitesMove, aDropPiece[nC, 1]);
            GenOnePieceMove(bWhitesMove, aDropPiece[nC, 4]);
        END;
    END
    ELSE
    IF (cMove[1] = 'R') THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 2]);
    END
    ELSE
    IF (cMove[1] = 'N') THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 3]);
    END
    ELSE
    IF (cMove[1] = 'B') THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 4]);
    END
    ELSE
    IF (cMove[1] = 'Q') THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 5]);
    END
    ELSE
    IF (cMove[1] = 'K') OR (CompareStr(cMove, 'O-O') = 0) OR (CompareStr(cMove, 'O-O-O') = 0) THEN
    BEGIN
        GenOnePieceMove(bWhitesMove, aDropPiece[nC, 6]);
    END
    ELSE
    BEGIN
        Result := 'Unknown piece move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + '].';
        EXIT;
    END;

    FOR nI := 1 TO nMove DO
    BEGIN
        IF (aMove[nI].nFromX = -1) THEN
            CONTINUE;

        BackupSetting;

        ExecMove(nI);

        IF (SquareAttacked(aKingBX[nC], aKingBY[nC], nC)) THEN
        BEGIN
            RestoreSetting;

            aMove[nI].nFromX := -1;
            aMove[nI].nFromY := -1;
            aMove[nI].nToX   := -1;
            aMove[nI].nToY   := -1;
        END
        ELSE
            RestoreSetting;
    END;

    nMI := 0;
    FOR nI := 1 TO nMove DO
    BEGIN
        IF (aMove[nI].nFromX = -1) THEN
            CONTINUE;

        GetAlgebraicMoveStr(nI, FALSE, FALSE, cM1, cM2, cM3, cM4);

        IF ((CompareStr(cMove, cM1) = 0) OR (CompareStr(cMove, cM2) = 0) OR
            (CompareStr(cMove, cM3) = 0) OR (CompareStr(cMove, cM4) = 0)) THEN
        BEGIN
            nMI := nI;
            BREAK;
        END;
    END;
    IF (nMI = 0) THEN
    BEGIN
        Result := 'Illegal move [' + IntToStr(nMoveNum) + cBoard + '. ' + cOrgMove + '].';
        EXIT;
    END;

    Result := '';
END;

FUNCTION TChess.GetRealMove(nMI: Integer; cOrgMove: String): String;
VAR
    cEnd: String;
    nS, nI, nJ: Integer;
BEGIN
    cEnd := '';
    nS   := Length(cOrgMove);
    nI   := 0;
    FOR nJ := nS DOWNTO 1 DO
    BEGIN
        IF ((cOrgMove[nJ] = '+') OR (cOrgMove[nJ] = '#')) THEN
            cEnd := cEnd + cOrgMove[nJ]
        ELSE
            BREAK;
    END;
    Result := GetMoveStr(nMI, FALSE, FALSE) + cEnd;
END;

procedure tchess.AddDropPiece(cpc:char);
var i,j:integer;
begin
 for i:=1 to 2 do
 for j:=6 to 8 do
 begin
  acpiece[i,j]:=EMPTY_SQUARE;
  anpiece[i,j]:=0;
 end;

 for i:=1 to 2 do
 for j:=1 to 5 do
 if adroppiece[i,j]=cpc then
 begin
  inc(anpiece[i,j]);
  if anpiece[i,j]=0 then acpiece[i,j]:=EMPTY_SQUARE
  else acpiece[i,j]:=cpc;
  exit;
 end;
end;

procedure tchess.goclear;
var i,j:integer;
begin
 for i:=1 to 8 do
 for j:=1 to 8 do
  if lowercase(aboard[i,j])<>'k' then
  aBoard[i,j]:=EMPTY_Square;
end;

procedure tchess.undoSetup;
begin
 setPosition(cSaveBoard,cSaveWPieces,cSaveBPieces);
end;

function tchess.getEnpassantsq:string;
var res,cfiles:string;
begin
 res:='';
 cfiles:='abcdefgh';
 if ((mprev.nFromY=2) and (mprev.nToY=4) and (mprev.cPiece=White_Pawn)) then
 begin
   res:=cfiles[mprev.nfromX]+'3';
 end;
 if ((mprev.nFromY=7) and (mprev.nToY=5) and (mprev.cPiece=Black_Pawn)) then
 begin
   res:=cfiles[mprev.nfromX]+'6';
 end;
 result:=res;
end;

function tchess.GetBFenStr:string;
var s:string;
begin
 s:=getFenBoardStr;
 result:=s;
end;




END. { CHESS }
